"""Config API.

See the `toolbox examples
<https://github.com/OHBA-analysis/osl-dynamics/tree/main/examples/toolbox_paper>`_
for scripts that use the config API.
"""

from osl_dynamics.config_api import pipeline, wrappers
