"""A collection of modules to perform post-hoc analysis of data.

"""

from osl_dynamics.analysis import (
    connectivity,
    gmm,
    modes,
    power,
    regression,
    spectral,
    static,
    statistics,
    workbench,
)
